import unittest
import os
import tempfile

from oplus.configuration import CONF
from oplus.simulation import simulate

DEBUG_SIMUL_DIR_PATH = None  # r"C:\Users\Geoffroy\Desktop\simul_dir"  # -> set to None to bypass


class TestValues(unittest.TestCase):
    def test_1ZoneUncontrolled(self):
        idf_path = os.path.join(CONF.eplus_base_dir_path, "ExampleFiles", "1ZoneUncontrolled.idf")
        epw_path = os.path.join(CONF.eplus_base_dir_path, "WeatherData", "USA_FL_Tampa.Intl.AP.722110_TMY3.epw")

        with tempfile.TemporaryDirectory() as temp_dir_path:
            s = simulate(idf_path, epw_path, temp_dir_path if DEBUG_SIMUL_DIR_PATH is None else DEBUG_SIMUL_DIR_PATH)
            eio = s.eio

        self.assertEqual(float(eio.df("Site:GroundReflectance:SnowModifier").loc[0, "Normal"]), 1)
        df = eio.df("Material CTF Summary")
        self.assertEqual(float(df[df[df.columns[0]] == "R13LAYER"].iloc[0, 5]), 2.291)
        self.assertEqual(float(eio.get_value("Material CTF Summary", 5, 0, "R13LAYER")), 2.291)
        self.assertEqual(float(eio.get_value("Material CTF Summary", "ThermalResistance {m2-K/w}", "Material Name",
                                             "R13LAYER")), 2.291)

        # print(eio("Construction CTF"))
        # print(eio("Material CTF Summary"))
        # print(eio("Zone Air Generic Contaminant Balance Simulation"))


import unittest
import os

from oplus.idf import BrokenIDFError, IsPointedError, IDF, IDFObject
from oplus.configuration import CONF


schedule_test_object_str = """Schedule:Compact,
    %s,  !- Name
    Any Number,              !- Schedule Type Limits Name
    THROUGH: 12/31,          !- Field 1
    FOR: AllDays,            !- Field 2
    UNTIL: 12:00,1,          !- Field 3
    UNTIL: 24:00,0;          !- Field 5"""


class OneZoneEvapCooler(unittest.TestCase):
    """
    Tested under EPlus 8.1.0 on Windows (Geoffroy).
    !!! Only tests that do not modify IDF (avoid loading idf several times) - else use OneZoneEvapCoolerDynamic.
    """
    @classmethod
    def setUpClass(cls):
        cls.idf_manager = IDF(os.path.join(CONF.eplus_base_dir_path, "ExampleFiles", "1ZoneEvapCooler.idf"))._

    def test_idf_call(self):
        qs = self.idf_manager.filter_by_ref("Construction")
        self.assertEqual({"R13WALL", "FLOOR", "ROOF31"}, set([c._.get_value("name") for c in qs]))

    def test_qs_one(self):
        self.assertEqual(self.idf_manager.filter_by_ref("BuildingSurface:Detailed"
                            ).filter("naMe", "Zn001:Roof001").one._.get_value("name"), "Zn001:Roof001")

    def test_idf_create_object(self):
        sch_name = "NEW TEST SCHEDULE"
        sch = self.idf_manager.add_object(schedule_test_object_str % sch_name)
        self.assertTrue(isinstance(sch, IDFObject))

    def test_pointing_links_l(self):
        zone = self.idf_manager.filter_by_ref("Zone").one
        d = {  # ref: [pointing_index, nb of objects], ...
            "BuildingSurface:Detailed": [3, 6],  # index 3
            "ZoneInfiltration:DesignFlowRate": [1, 1],  # index 1
            "ZoneHVAC:EquipmentConnections": [0, 1],  # index 0
            "ZoneControl:Thermostat": [1, 1]  # index 1
        }

        # check all pointing
        _d = {}
        for pointing_object, pointing_index in zone._.get_pointing_links_l():
            # check points
            self.assertEqual(pointing_object._.get_value(pointing_index), zone)
            # verify all are identified
            if not pointing_object._.ref in _d:
                _d[pointing_object._.ref] = [pointing_index, 1]
            else:
                self.assertEqual(pointing_index, _d[pointing_object._.ref][0])
                _d[pointing_object._.ref][1] += 1
        self.assertEqual(d, _d)

        # check pointing on pointed_index
        self.assertEqual(len(zone._.get_pointing_links_l(0)), 9)  # 9 pointing
        self.assertEqual(len(zone._.get_pointing_links_l(3)), 0)  # no pointing


class OneZoneEvapCoolerDynamic(unittest.TestCase):
    """
    Tested under EPlus 8.1.0 on Windows (Geoffroy).
    Here are tests that modify idf.
    """
    def setUp(self):
        self.idf_manager = IDF(os.path.join(CONF.eplus_base_dir_path, "ExampleFiles", "1ZoneEvapCooler.idf"))._

    def test_idf_add_object(self):
        sch_name = "NEW TEST SCHEDULE"
        new_object = self.idf_manager.add_object(schedule_test_object_str % sch_name)
        self.assertEqual(new_object._.idf_manager, self.idf_manager)

    def test_idf_add_object_broken(self):
        self.assertRaises(BrokenIDFError, lambda: self.idf_manager.add_object("""
        Material,
        C5 - 4 IN HW CONCRETE,   !- Name
        MediumRough,             !- Roughness
        0.1014984,               !- Thickness {m}
        1.729577,                !- Conductivity {W/m-K}
        2242.585,                !- Density {kg/m3}
        836.8000,                !- Specific Heat {J/kg-K}
        0.9000000,               !- Thermal Absorptance
        0.6500000,               !- Solar Absorptance
        0.6500000;               !- Visible Absorptance
        """))

    def test_idf_remove_object(self):
        sch_name = "NEW TEST SCHEDULE"
        sch = self.idf_manager.add_object(schedule_test_object_str % sch_name)
        self.idf_manager.remove_object(sch)
        self.assertEqual(len(self.idf_manager.filter_by_ref("Schedule:Compact").filter("name", sch_name)), 0)

    def test_idf_remove_object_raise(self):
        zone = self.idf_manager.filter_by_ref("Zone").one
        self.assertRaises(IsPointedError, lambda: self.idf_manager.remove_object(zone, raise_if_pointed=True))

    def test_idf_remove_object_no_raise(self):
        zone = self.idf_manager.filter_by_ref("Zone").one
        pointing_links_l = zone._.get_pointing_links_l()
        self.idf_manager.remove_object(zone, raise_if_pointed=False)
        # check that pointing's pointed fields have been removed
        for pointing_object, pointing_index in pointing_links_l:
            self.assertEqual(pointing_object._.get_value(pointing_index), None)

    def test_set_value_object(self):
        new_name = "Fan Availability Schedule - 2"
        supply_fan = self.idf_manager.filter_by_ref("Fan:ConstantVolume").filter("name", "Supply Fan").one
        supply_fan._.set_value("availability schedule name", schedule_test_object_str % new_name)
        # check set
        self.assertEqual(
            self.idf_manager.filter_by_ref("Fan:ConstantVolume").filter(
                "name", "Supply Fan").one._.get_value("AvaiLABIlity schedule name")._.get_value(
                "NAME"), new_name)

    def test_set_value_reference(self):
        new_zone_name = "new zone name"
        zone = self.idf_manager.filter_by_ref("Zone").one
        pointing_links_l = zone._.get_pointing_links_l()
        zone._.set_value("name", new_zone_name)
        # print(zone)
        self.assertEqual(zone._.get_value("name"), new_zone_name)
        for pointing_object, pointing_index in pointing_links_l:
            self.assertEqual(pointing_object._.get_raw_value(pointing_index), new_zone_name)

    def test_copy_object(self):
        zone = self.idf_manager.filter_by_ref("Zone").one
        new = zone._.copy()
        for i in range(zone._.fields_nb):
            if i == 0:
                self.assertNotEqual(zone._.get_value(i), new._.get_value(i))
            else:
                self.assertEqual(zone._.get_value(i), new._.get_value(i))

    def test_replace_values(self):
        sch = self.idf_manager.filter_by_ref("Schedule:Compact").filter("name", "Heating Setpoint Schedule").one
        pointing_l = [o for (o, i) in sch._.get_pointing_links_l()]

        # bigger
        new_str = """Schedule:Compact,
My New Heating Setpoint Schedule,  !- Name
Any Number,              !- Schedule Type Limits Name
Through: 12/31,          !- Field 1
For: AllDays,            !- Field 2
Until: 12:00,20.0,       !- Field 3
Until: 24:00,25.0;       !- Field 3"""
        sch._.replace_values(new_str)

        self.assertEqual(sch["name"], "Heating Setpoint Schedule")
        self.assertEqual([o for (o, i) in sch._.get_pointing_links_l()], pointing_l)

        # smaller
        new_str = """Schedule:Compact,
,  !- Name
Any Number,              !- Schedule Type Limits Name
Through: 12/31,          !- Field 1
For: AllDays,            !- Field 2
Until: 12:00,20.0;       !- Field 3"""

        sch._.replace_values(new_str)
        self.assertEqual(len(sch), 6)



"""
We respect private/public naming conventions for methods and variables, EXCEPT for IDF or IDFObject managers. The
_manager variable is semi-private: it can be accessed by other managers (including other modules of oplus), but not by
IDF or IDFObject. The _manager attributes therefore remain private to oplus users.
"""
# todo: document

import io
import datetime as dt
import os


from oplus.configuration import CONF
from oplus.idd import IDD
from oplus.util import get_copyright_comment, Cached, check_cache_is_off, cached, get_string_buffer


class IDFError(Exception):
    pass


class BrokenIDFError(IDFError):
    pass


class IsPointedError(IDFError):
    pass


class ObjectDoesNotExist(IDFError):
    pass


class MultipleObjectsReturned(IDFError):
    pass


# ----------------------------------------------------------------------------------------------------------------------
# --------------------------------------------- IDF OBJECTS ------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------


# ----------------------------------------------- idf object  ----------------------------------------------------------
class IDFObject:
    """
    IDFObject is allowed to access private keys/methods of IDF.
    """
    def __init__(self, object_manager):
        self._ = object_manager

    def __str__(self):
        """pretty prints"""
        return self._.to_str(style="console")

    def __repr__(self):
        """concise print"""
        return "<Idf object: %s>" % self.ref

    def __getitem__(self, item):
        """
        Arguments
        ---------
        item: field index or name [or slice of field indexes and/or names]
        """
        if isinstance(item, slice):
            start = 0 if item.start is None else self._.get_field_index(item.start)
            stop = len(self) if item.stop is None else min(self._.get_field_index(item.stop), len(self))
            step = item.step or 1
            return [self._.get_value(i) for i in range(start, stop, step)]
        return self._.get_value(item)

    def get(self, item, default=None):
        try:
            return self[item]
        except (AttributeError, IndexError):
            return default

    def __setitem__(self, key, value):
        """
        Arguments
        ---------
        key: field index or name
        value: field raw, parsed or object value (see get_value documentation)
        """
        self._.set_value(key, value)

    def __len__(self):
        return self._.fields_nb

    def __iter__(self):
        """
        Iter through fields of object.
        """
        def iter_var():
            for i in range(len(self)):
                yield self[i]
        return iter_var()

    @property
    def idf(self):
        return self._.idf_manager.idf

    @property
    def ref(self):
        """
        Object descriptor ref
        """
        return self._.ref

    @property
    def pointing_objects(self):
        return self._.pointing_objects

    def to_str(self, style="idf"):
        return self._.to_str(style=style)

    def info(self, detailed=False):
        """
        Returns a string with all available fields of object (information provided by the idd).

        Arguments
        ---------
            detailed: include all field tags information
        """
        return self._.info(detailed=detailed)

    def copy(self):
        return self._.copy()

    def add_field(self, raw_value_or_value, comment=""):
        """
        Add a new field to object (at the end).

        Arguments
        ---------
        rawvalue_or_value: see get_value for mor information
        comment: associated comment
        """
        self._.add_field("", comment=comment)
        self._.set_value(self._.fields_nb-1, raw_value_or_value)

    def replace_values(self, new_object_str):
        """
        Replaces all values of object that are not links (neither pointing nor pointed fields) with values contained
        in the idf object string 'new_object_str'.
        """
        self._.replace_values(new_object_str)

    def pop(self, index=-1):
        """
        Removes field from idf object and shift following rows upwards (value and comment will be removed).
        Can only be applied on extensible fields (for now, only extensible:1).

        #todo: manage extensible > 1

        Arguments
        ---------
        index: index of field to remove (default -1)

        Returns
        -------
        Value of poped field.
        """
        return self._.pop(index)

    def field_comment(self, field_index_or_name, comment=None):
        if comment is None:
            return self._.get_field_comment(field_index_or_name)
        self._.set_field_comment(field_index_or_name, comment)

    @property
    def head_comment(self):
        return self._.get_head_comment()

    @head_comment.setter
    def head_comment(self, value):
        self._.set_head_comment(value)

    @property
    def tail_comment(self):
        return self._.get_tail_comment()

    @tail_comment.setter
    def tail_comment(self, value):
        self._.set_tail_comment(value)


# ------------------------------------------- idf object manager -------------------------------------------------------
class IDFObjectManager(Cached):
    idf_object_cls = IDFObject  # for subclassing

    _COMMENT_COLUMN_START = 35
    _TAB_LEN = 4
    _ROWS_NB = 120

    _RAW_VALUE = 0
    _COMMENT = 1

    def __init__(self, ref, idf_manager, head_comment=None, tail_comment=None):
        self._ref = ref
        self._head_comment = head_comment
        self._tail_comment = tail_comment
        self._fields_l = []  # [[raw_value, comment], ...]

        self._idf_manager = idf_manager
        self._descriptor = self._idf_manager.idd.get_object_descriptor(ref)

        self._idf_object = self.idf_object_cls(self)

    # ---------------------------------------------- EXPOSE ------------------------------------------------------------
    @property
    def ref(self):
        return self._ref

    @property
    def idf_object(self):
        return self._idf_object

    @property
    def fields_nb(self):
        return len(self._fields_l)

    @property
    def idf_manager(self):
        return self._idf_manager

    # --------------------------------------------- CONSTRUCT ----------------------------------------------------------
    @check_cache_is_off
    def add_field(self, raw_value, comment=""):
        if not isinstance(raw_value, str):
            raise IDFError("'raw_value' must be a string.")
        self._fields_l.append([raw_value, comment])

    @check_cache_is_off
    def add_tail_comment(self, comment):
        stripped = comment.strip()
        if stripped == "":
            return None
        if self._tail_comment is None:
            self._tail_comment = ""
        self._tail_comment += "%s\n" % stripped

    def copy(self):
        # create new object
        new_object_manager = self._idf_manager.idf_object_manager_cls(self._ref, self._idf_manager,
                                                                      head_comment=self._head_comment,
                                                                      tail_comment=self._tail_comment)
        # we must change all references
        for i in range(len(self._fields_l)):
            fieldd = self._descriptor.get_field_descriptor(i)
            if fieldd.detailed_type == "reference":
                raw_value = dt.datetime.now().strftime("%Y%m%d%H%M%S%f") + str(i)
            else:
                raw_value = self.get_raw_value(i)
            new_object_manager.add_field(raw_value, comment=self.get_field_comment(i))

        # add object to idf
        return self._idf_manager.add_object_from_parsed(new_object_manager)

    # ---------------------------------------------- DESTROY -----------------------------------------------------------
    @check_cache_is_off
    def remove_values_that_point(self, pointed_object=None):
        """
        Removes all fields that point at pointed_object if not None (if you want a specific index, use remove_value).
        """
        for i in range(len(self._fields_l)):
            fieldd = self._descriptor.get_field_descriptor(i)
            if fieldd.detailed_type == "object-list":
                if (pointed_object is None) or (self.get_value(i) is pointed_object):
                    self.set_value(i, None)

    @check_cache_is_off
    def neutralize(self):
        """
        remove values and links of fields, idf_manager, descriptor, pointing_d.
        """
        self._idf_manager = None
        self._descriptor = None

    @check_cache_is_off
    def pop(self, field_index_or_name=-1):
        """
        Remove item at given position and return it. All rows will be shifted upwards to fill the blank. May only be
        used on extensible fields.
        For the moment, only extensible=1 is coded.
        """
        # check extensible
        extensible_cycle_len, extensible_start_index = self._descriptor.extensible
        if extensible_cycle_len != 1:
            raise IDFError("Can only use pop on fields defined as 'extensible:1'.")

        # check index
        pop_index = self.get_field_index(field_index_or_name)
        if pop_index < extensible_start_index:
            raise IDFError("Can only use pop on extensible fields (pop index must be > than extensible start index).")

        # store value
        pop_value = self.get_value(pop_index)

        # remove field
        self._fields_l.pop(pop_index)

        return pop_value

    # ------------------------------------------------ GET -------------------------------------------------------------
    def get_field_index(self, field_index_or_name):
        """
        Returns field index (>=0).
        """
        if isinstance(field_index_or_name, int):
            field_index_or_name = (field_index_or_name if field_index_or_name >= 0 else
                                   self.fields_nb + field_index_or_name)
        return self._descriptor.get_field_index(field_index_or_name)

    def get_raw_value(self, field_index_or_name):
        field_index = self.get_field_index(field_index_or_name)
        return self._fields_l[field_index][self._RAW_VALUE]

    @cached
    def get_value(self, field_index_or_name):
        # parsed value and/or object
        index = self.get_field_index(field_index_or_name)
        raw_value = self._fields_l[index][self._RAW_VALUE]
        fieldd = self._descriptor.get_field_descriptor(index)
        if fieldd.detailed_type in ("integer", "real", "alpha", "choice", "node", "reference", "external-list"):
            value, pointed_index = fieldd.basic_parse(raw_value), None
        elif fieldd.detailed_type == "object-list":
            value, pointed_index = self._idf_manager.get_pointed_link(self._ref, index, raw_value)
        else:
            raise NotImplementedError("Unknown field type : '%s'." % fieldd.detailed_type)
        return value

    def get_field_comment(self, field_index_or_name):
        field_index = self.get_field_index(field_index_or_name)
        comment = self._fields_l[field_index][self._COMMENT]
        return "" if comment is None else comment

    def get_head_comment(self):
        return "" if self._head_comment is None else self._head_comment

    def get_tail_comment(self):
        return "" if self._tail_comment is None else self._tail_comment

    @cached
    def get_pointing_links_l(self, field_index_or_name=None):
        index_l = (range(len(self._fields_l)) if field_index_or_name is None
                   else [self.get_field_index(field_index_or_name)])

        links_l = []
        for i in index_l:
            fieldd = self._descriptor.get_field_descriptor(i)
            if fieldd.detailed_type != "reference":
                continue
            links_l.extend(self.idf_manager.get_pointing_links_l(self._ref, i, self.get_raw_value(i)))

        return links_l

    @property
    @cached
    def pointing_objects(self):
        return QuerySet([pointing_object for pointing_object, pointing_index in self.get_pointing_links_l()])

    # def get_pointed_links_l(self, field_index_or_name=None):
    #     """
    #     not used, not tested
    #     """
    #     index_l = (range(len(self._fields_l)) if field_index_or_name is None
    #                else [self.get_field_index(field_index_or_name)])
    #     links_l = []
    #     for i in index_l:
    #         value = self._fields_l[i][self._VALUE]
    #         if isinstance(value, IDFObject):
    #             links_l.append((value, self._fields_l[i][self._POINTED_INDEX]))
    #
    #     return links_l

    # ------------------------------------------------ SET -------------------------------------------------------------
    @check_cache_is_off
    def set_value(self, field_index_or_name, raw_value_or_value, raise_if_pointed=False):
        field_index = self.get_field_index(field_index_or_name)
        fieldd = self._descriptor.get_field_descriptor(field_index)

        if fieldd.detailed_type in ("integer", "real", "alpha", "choice", "node"):  # basic type
            # store and parse
            raw_value = "" if raw_value_or_value is None else str(raw_value_or_value).strip()
            self._fields_l[field_index][self._RAW_VALUE] = raw_value
            # self._parse_field(field_index)

        elif fieldd.detailed_type == "reference":
            if raise_if_pointed:
                if len(self.get_pointing_links_l(field_index)) > 0:
                    raise IsPointedError("Set field is already pointed by another object. First remove this object, "
                                         "or disable 'raise_if_pointed' argument.")
            # remove all pointing
            pointing_links_l = self.get_pointing_links_l(field_index)
            for pointing_object, pointing_index in pointing_links_l:
                pointing_object._.set_value(pointing_index, None)
            # store and parse
            raw_value = "" if raw_value_or_value is None else str(raw_value_or_value).strip()
            self._fields_l[field_index][self._RAW_VALUE] = raw_value
            # self._parse_field(field_index)
            # re-set all pointing
            for pointing_object, pointing_index in pointing_links_l:
                pointing_object._.set_value(pointing_index, self._idf_object)

        elif fieldd.detailed_type == "object-list":  # detailed type
            # convert to object if necessary
            if raw_value_or_value is None:
                self._fields_l[field_index][self._RAW_VALUE] = ""
            else:
                if isinstance(raw_value_or_value, str):
                    try:
                        value = self._idf_manager.add_object(raw_value_or_value)
                    except BrokenIDFError as e:
                        raise e
                    except Exception as e:
                        raise IDFError("Error while parsing string to create new object. Give object if "
                                       "exists or correct idf string.\n\n%s" % e)

                elif isinstance(raw_value_or_value, IDFObject):
                    value = raw_value_or_value
                else:
                    raise IDFError("Wrong value descriptor: '%s' (instead of IDFObject)." % type(raw_value_or_value))

                # check if correct idf
                if not value._.idf_manager is self._idf_manager:
                    raise IDFError("Given object does not belong to correct idf: '%s' instead of '%s." %
                                   (value._.idf_manager.path, self._idf_manager.path))

                # check that new object ref can be set, and find pointed index
                pointed_index = None
                for link_name in fieldd.get_tag("object-list"):
                    # link: (object_ref, index)
                    if pointed_index is not None:
                        break
                    for object_descriptor, object_index in self._idf_manager.idd.pointed_links(link_name):
                        if ((value._.ref == object_descriptor.ref) and
                                (value._.get_value(object_index) is not None)):
                            # ok, we fond an accepted combination
                            pointed_index = object_index
                            break
                if pointed_index is None:
                    raise IDFError("Wrong value ref: '%s' for field '%i' of object descriptor ref '%s'. "
                                   "Can't set object." % (value._.ref, field_index, self._ref))
                # get raw value
                raw_value = value._.get_raw_value(pointed_index)

                # now we checked everything was ok, we remove old field (so pointed object unlinks correctly)
                self._fields_l[field_index][self._RAW_VALUE] = ""

                # store and parse
                self._fields_l[field_index][self._RAW_VALUE] = raw_value
        else:
            raise IDFError("Unknown detailed type: '%s'." % fieldd.detailed_type)

        # remove if last field and emptied
        if (len(self._fields_l) == (field_index+1)) and self._fields_l[-1][self._RAW_VALUE] == "":
            self._fields_l.pop()

    @check_cache_is_off
    def replace_values(self, new_object_str):
        """
        Purpose: keep old pointed links
        """
        # create object (but it will not be linked to idf)
        objects_l, comments = self._idf_manager.parse(io.StringIO(new_object_str))  # comments not used
        if len(objects_l) != 1:
            raise IDFError("Wrong number of objects created: %i" % len(objects_l))
        new_object = objects_l[0]

        if self.ref != new_object._.ref:
            raise IDFError("New object (%s) does not have same reference as new object (%s). Can't replace." %
                           (self.ref, new_object._.ref))

        # replace fields using raw_values, one by one
        old_nb, new_nb = self.fields_nb, new_object._.fields_nb

        for i in range(max(old_nb, new_nb)):
            fieldd = self._descriptor.get_field_descriptor(i)
            if fieldd.detailed_type in ("reference", "object-list"):
                continue  # we do not modifiy links
            if (i < old_nb) and (i < new_nb):
                self.set_value(i, new_object._.get_raw_value(i))
            elif i < old_nb:
                self.set_value(i, None)
            else:
                self.add_field(new_object._.get_raw_value(i), new_object._.get_field_comment(i))

        # remove all last values
        self._fields_l = self._fields_l[:new_nb]

    @check_cache_is_off
    def set_field_comment(self, field_index_or_name, comment):
        field_index = self.get_field_index(field_index_or_name)
        comment = None if comment.strip() == "" else str(comment).replace("\n", " ").strip()
        self._fields_l[field_index][self._COMMENT] = comment

    @check_cache_is_off
    def set_head_comment(self, comment):
        """
        All line return will be replaced by a blank.
        """
        comment = None if comment.strip() == "" else str(comment).replace("\n", " ").strip()
        self._head_comment = comment

    @check_cache_is_off
    def set_tail_comment(self, comment):
        comment = None if comment.strip() == "" else str(comment).strip()
        self._tail_comment = comment

    # ------------------------------------------------ COMMUNICATE -----------------------------------------------------
    def to_str(self, style="idf"):
        if style == "idf":
            # object descriptor ref
            content = "%s" % self._descriptor.ref + ("," if self.fields_nb != 0 else ";")
            spaces_nb = self._COMMENT_COLUMN_START - len(content)
            if spaces_nb < 0:
                spaces_nb = self._TAB_LEN
            comment = "" if self._head_comment is None else "%s! %s" % (" "*spaces_nb, self._head_comment)
            s = content + comment + "\n"

            # fields
            for field_index, (f_raw_value, f_comment) in enumerate(self._fields_l):
                content = "%s%s%s" % (" " * self._TAB_LEN,
                                      f_raw_value,
                                      ";" if field_index == len(self._fields_l)-1 else ",")
                spaces_nb = self._COMMENT_COLUMN_START - len(content)
                if spaces_nb < 0:
                    spaces_nb = self._TAB_LEN
                comment = "" if f_comment is None else "%s! %s" % (" "*spaces_nb, f_comment)
                s += content + comment + "\n"

            # tail comment
            if self._tail_comment is not None:
                s += "\n"
                for line in self._tail_comment.strip().split("\n"):
                    s += "! %s\n" % line

        elif style == "console":
            s = "%s\n%s%s\n" % ("-" * self._ROWS_NB, str(self.to_str(style="idf")), "-" * self._ROWS_NB)
        else:
            raise IDFError("Unknown style: '%s'." % style)

        return s

    def info(self, detailed=False):
        """
        Returns a string with all available fields of object (information provided by the idd).

        Arguments
        ---------
            detailed: include all field tags information
        """
        msg = "%s\n%s\n%s" % ("-"*len(self._ref), self._ref, "-"*len(self._ref))
        for i, fd in enumerate(self._descriptor.field_descriptors_l):
            msg += "\n%i: %s" % (i, fd.name)
            if detailed:
                for tag in fd.tags:
                    msg += "\n\t* %s: %s" % (tag, fd.get_tag(tag))
        return msg


# ----------------------------------------------------------------------------------------------------------------------
# ------------------------------------------------- IDF ----------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------- idf manager  ---------------------------------------------------------
class VoidSimulation:
    def __getattr__(self, item):
        raise IDFError("Idf is not attached to a simulation, '%s' enhancement is not available." % item)


class IDFManager(Cached):
    idf_object_manager_cls = IDFObjectManager  # for subclassing

    # ----------------------------------------------- INITIALIZE -------------------------------------------------------
    def __init__(self, idf, path_or_content, idd_or_path=None, encoding=None):
        self._idf = idf
        self._idd = IDD.get_idd(idd_or_path, encoding=encoding)
        self._encoding = CONF.encoding if encoding is None else encoding

        # simulation
        self._simulation = VoidSimulation()  # must be before parsing

        # get string buffer and store path (for info)
        buffer, path = get_string_buffer(path_or_content, "idf", self._encoding)
        self._path = path_or_content

        # raw parse and parse
        with buffer as f:
            try:
                self._objects_l, self._head_comments = self.parse(f)
            except Exception as e:
                raise IDFError("Error while parsing idf. First check that given file exists "
                               "(if not, given path will have been considered as an idf content.\n%s" % e)

    # ----------------------------------------------- EXPOSE -----------------------------------------------------------
    @staticmethod
    def copyright_comment():
        return get_copyright_comment()

    @property
    def objects_l(self):
        return self._objects_l

    @property
    def idd(self):
        return self._idd

    @property
    def simulation(self):
        return self._simulation

    @property
    def idf(self):
        return self._idf

    # --------------------------------------------- CONSTRUCT ----------------------------------------------------------
    def parse(self, file_like):
        """
        Objects are created from string. They are not attached to idf manager yet.
        """
        objects_l, head_comments = [], ""
        idf_object_manager = None
        make_new_object = True
        for i, raw_line in enumerate(file_like):
            # GET LINE CONTENT AND COMMENT
            split_line = raw_line.split("!")
            if len(split_line) == 0:
                content, comment = None, None
            elif len(split_line) == 1:
                if "!" in raw_line:
                    content, comment = None, split_line[0]
                else:
                    content, comment = split_line[0].strip(), None
            else:
                content, comment = split_line[0].strip(), "!".join(split_line[1:])

            # RAW FORMATTING
            if content.strip() == "":
                content = None
            if comment is not None:
                comment = comment.strip()
                if comment == "":
                    comment = None

            # SKIP CURRENT LINE IF VOID
            if (content, comment) == (None, None):
                continue

            # NO CONTENT
            if content is None:
                if idf_object_manager is None:  # head idf comment
                    head_comments += "\n%s" % comment
                else:
                    idf_object_manager.add_tail_comment(comment)
                continue

            # CONTENT
            # check if object end and prepare
            object_end = content[-1] == ";"
            content = content[:-1]  # we tear comma or semi-colon
            content_l = [text.strip() for text in content.split(",")]

            # object creation if needed
            if make_new_object:
                idf_object_manager = self.idf_object_manager_cls(content_l[0].strip(), self, head_comment=comment)
                objects_l.append(idf_object_manager.idf_object)
                # prepare in case fields on the same line
                content_l = content_l[1:]
                make_new_object = False

            # fields
            for value_s in content_l:
                idf_object_manager.add_field(value_s, comment=comment)

            # signal that new object must be created
            if object_end:
                make_new_object = True

        return objects_l, head_comments

    # ----------------------------------------------- LINKS ------------------------------------------------------------
    @cached
    def get_pointed_link(self, pointing_ref, pointing_index, pointing_raw_value):
        # get field descriptor
        fieldd = self._idd.get_object_descriptor(pointing_ref).get_field_descriptor(pointing_index)
        # check if object-list
        if fieldd.detailed_type != "object-list":
            raise IDFError("Only 'object-list' fields can point on an object. Wrong field given. Ref: '%s', "
                           "index: '%i'." % (pointing_ref, pointing_index))
        # check if an object is pointed
        if pointing_raw_value == "":  # no object pointed
            return None, None
        # iter through link possibilities and return if found
        link_names_l = fieldd.get_tag("object-list")
        for link_name in link_names_l:
            for od, field_index in self._idd.pointed_links(link_name):
                for idf_object in self.filter_by_ref(od.ref):
                    if idf_object._.get_raw_value(field_index) == pointing_raw_value:
                        return idf_object, field_index
        raise IDFError("Link not found. Field 'object-list' tag values: %s, field value : '%s'" %
                       (str(link_names_l), pointing_raw_value))

    @cached
    def get_pointing_links_l(self, pointed_ref, pointed_index, pointed_raw_value):
        # get field descriptor
        fieldd = self.idd.get_object_descriptor(pointed_ref).get_field_descriptor(pointed_index)
        # check if reference
        if fieldd.detailed_type != "reference":
            raise IDFError("Only 'reference' fields can be pointed by an object. Wrong field given. Ref: '%s', "
                           "index: '%i'." % (pointed_ref, pointed_index))
        # check if an object can be pointing
        if pointed_raw_value == "":
            return []
        # fetch links
        links_l = []
        for link_name in fieldd.get_tag("reference"):
            for object_descriptor, pointing_index in self.idd.pointing_links(link_name):
                for idf_object in self.filter_by_ref(object_descriptor.ref):
                    if pointing_index >= idf_object._.fields_nb:
                        continue
                    if idf_object._.get_raw_value(pointing_index) == pointed_raw_value:
                        links_l.append([idf_object, pointing_index])
        return links_l

    def check_new_reference(self, new_object_ref, new_object_index, reference):
        if reference == "":
            return None

        object_descriptor = self._idd.get_object_descriptor(new_object_ref)
        fieldd = object_descriptor.get_field_descriptor(new_object_index)

        # iter through all possible links and check that reference is not already in use
        for link_name in fieldd.get_tag("reference"):
            for pointed_descriptor, pointed_index in self._idd.pointed_links(link_name):
                for pointed_object in self.filter_by_ref(pointed_descriptor.ref):
                    if pointed_object._.get_raw_value(pointed_index) == reference:
                        raise BrokenIDFError("New object has same reference at index '%s' as other object of "
                                             "same link name. Other object ref: '%s', index: '%s'. The value at"
                                             " that field must be changed." %
                                             (new_object_index, pointed_object._.ref, pointed_index))

    # ------------------------------------------ MANAGE OBJECTS --------------------------------------------------------
    def has_object(self, idf_object):
        return idf_object in self._objects_l

    @check_cache_is_off
    def add_object(self, new_str, position=None):
        """
        From str
        """
        # create object
        objects_l, comments_l = self.parse(io.StringIO(new_str))  # comments not used (only for global idf parse)
        if len(objects_l) != 1:
            raise IDFError("Wrong number of objects created: %i" % len(objects_l))
        new_object = objects_l[0]
        return self.add_object_from_parsed(new_object._, position=position)

    @check_cache_is_off
    def add_object_from_parsed(self, raw_parsed_object_manager, position=None):
        """checks references uniqueness"""
        new_object = raw_parsed_object_manager.idf_object  # change name since no more raw parsed

        # check reference uniqueness
        object_descriptor = self._idd.get_object_descriptor(new_object._.ref)
        for i in range(new_object._.fields_nb):
            fieldd = object_descriptor.get_field_descriptor(i)
            if fieldd.detailed_type == "reference":
                self.check_new_reference(object_descriptor.ref, i, new_object._.get_raw_value(i))

        # add object
        if position is None:
            self._objects_l.append(new_object)
        else:
            self._objects_l.insert(position, new_object)

        # return new object
        return new_object

    @check_cache_is_off
    def remove_object(self, idf_object, raise_if_pointed=True):
        """
        Arguments
        ---------
        raise_if_pointed: raises Exception if is pointed by other objects.
            Else, sets all pointing object fields to None.
        """
        # check if object is pointed, if asked
        pointing_links_l = idf_object._.get_pointing_links_l()
        if raise_if_pointed and len(pointing_links_l) > 0:
            raise IsPointedError("Can't remove object if other objects are pointing to it and 'check' is "
                                 "True. Pointing objects: '%s'" % [o for (o, i) in pointing_links_l])

        # remove from pointing
        for pointing_object, pointing_index in pointing_links_l:
            pointing_object._.remove_values_that_point(idf_object)

        # remove pointed
        idf_object._.remove_values_that_point()

        # delete obsolete attributes
        idf_object._.neutralize()

        # remove from idf
        index = self._objects_l.index(idf_object)
        del self._objects_l[index]

        return index

    @cached
    def filter_by_ref(self, ref=None):
        if ref is None:
            return QuerySet(self.objects_l)
        return QuerySet(self._objects_l)(ref)

    # ------------------------------------------ MANAGE COMMENTS -------------------------------------------------------
    def get_comment(self):
        return self._head_comments

    @check_cache_is_off
    def set_comment(self, value):
        self._head_comments = str(value).strip()

    # ------------------------------------------------ COMMUNICATE -----------------------------------------------------
    def info(self, sort_by_group=False, detailed=False):
        """
        Indicates all objects references contained in current idf.
        Arguments
        ---------
        sort_by_group: will sort object descriptors by group
        detailed: will give all object descriptors' associated tags

        Returns
        -------
        a text describing the information on object contained in idd file
        """
        # ods: object descriptors
        def _get_ods_info(_ods, _line_start=""):
            _msg = ""
            for _od in sorted(_ods, key=lambda x: x.ref):
                _msg += "\n%s%s" % (_line_start, _od.ref)
                if detailed:
                    for _tag in _od.tags:
                        _msg += "\n%s\t* %s: %s" % (_line_start, _tag, _od.get_tag(_tag))
            return _msg

        ods_refs_set = set([idf_object.ref for idf_object in self._objects_l])
        name = "IDF: '%s'" % self._path
        msg = "%s\n%s\n%s" % ("-"*len(name), name, "-"*len(name))
        if sort_by_group:
            for group_name in self._idd.groups_l:
                ods_l = []
                for od in self._idd.get_object_descriptors_by_group(group_name):
                    if od.ref in ods_refs_set:
                        ods_l.append(od)
                if len(ods_l) > 0:
                    msg += "\nGroup - %s" % group_name
                    msg += _get_ods_info(ods_l, _line_start="\t")
        else:
            msg += _get_ods_info([self._idd.get_object_descriptor(od_ref) for od_ref in ods_refs_set])

        return msg

    def to_str(self, add_copyright=True):
        content = ""

        # idf comments
        idf_comment = self._head_comments
        if add_copyright:
            msg = self.copyright_comment()
            if msg not in idf_comment:
                idf_comment = msg + idf_comment

        for comment in idf_comment.split("\n"):
            content += "!%s\n" % comment

        # idf objects
        for idf_object in self._objects_l:
            content += "\n%s" % idf_object._.to_str(style="idf")

        return content

    def save_as(self, file_or_path, add_copyright=True):
        is_path = isinstance(file_or_path, str)
        f = open(file_or_path, "w", encoding=self._encoding) if is_path else file_or_path
        f.write(self.to_str(add_copyright=add_copyright))
        if is_path:
            f.close()

    def copy(self, add_copyright=True):
        content = self.to_str(add_copyright=add_copyright)
        return self.idf.__class__(content, self.idd, encoding=self._encoding)


# ------------------------------------------------- idf ----------------------------------------------------------------
class IDF:
    """
    IDF is allowed to access private keys/methods of IDFObject.
    """
    idf_manager_cls = IDFManager  # for subclassing

    @classmethod
    def get_idf(cls, idf_or_path, encoding=None):
        """
        Arguments
        ---------
        idf_or_path: idf object or idf file path

        Returns
        -------
        IDF object
        """
        if isinstance(idf_or_path, str):
            return cls(idf_or_path, encoding=encoding)
        elif isinstance(idf_or_path, cls):
            return idf_or_path
        raise IDFError("'idf_or_path' must be a path or an IDF. Given object: '%s', type: '%s'." %
                       (idf_or_path, type(idf_or_path)))

    def __init__(self, path_or_content, idd_or_path=None, encoding=None):
        """
        Arguments
        ---------
        path_or_content: idf path, content str, content bts or file_like. If path, must end by .idf.
        idd_or_path: IDD object or idd path. If None, default will be chosen (most recent EPlus version installed on
            computer)
        """
        self._ = self.idf_manager_cls(self, path_or_content, idd_or_path=idd_or_path, encoding=encoding)

    def __call__(self, object_descriptor_ref=None):
        """returns all objects of given object descriptor"""
        return self._.filter_by_ref(object_descriptor_ref)

    def to_str(self, add_copyright=True):
        return self._.to_str(add_copyright=add_copyright)

    def save_as(self, file_or_path):
        self._.save_as(file_or_path)

    def copy(self, add_copyright=True):
        return self._.copy(add_copyright=add_copyright)

    def remove_object(self, object_to_remove, raise_if_pointed=True):
        """
        Removes object from idf.

        Arguments
        ---------
        old: object to remove
        check: check if links have been broken. If check is True and broken links are detected, will raise an IDFError.
            (nodes or branches checking has not been implemented)
        """
        return self._.remove_object(object_to_remove, raise_if_pointed=raise_if_pointed)

    def add_object(self, new_str, position=None):
        """
        Adds new object to the idf, at required position.

        Arguments
        ---------
        new_or_str: new object (or string describing new object) that will be added to idf
        position: if None, will be added at the end, else will be added at asked position
            (using 'insert' python builtin function for lists)
        check: check if pointed objects of new object exists. If check is True and a non existing link is detected, will
            raise an IDFError
        """
        return self._.add_object(new_str, position=position)

    def info(self, sort_by_group=False, detailed=False):
        """
        Arguments
        ---------
        sort_by_group: will sort object descriptors by group
        detailed: will give all object descriptors' associated tags

        Returns
        -------
        a text describing the information on object contained in idd file
        """
        return self._.info(sort_by_group=sort_by_group, detailed=detailed)

    @property
    def comment(self):
        return self._.get_comment()

    @comment.setter
    def comment(self, value):
        self._.set_comment(value)

    def activate_cache(self):
        self._.activate_cache()
        for o in self._.objects_l:
            o._.activate_cache()

    def deactivate_cache(self):
        self._.deactivate_cache()
        for o in self._.objects_l:
            o._.deactivate_cache()

    def clear_cache(self):
        self._.clear_cache()
        for o in self._.objects_l:
            o._.clear_cache()


# ----------------------------------------------------------------------------------------------------------------------
# --------------------------------------------- QuerySet ---------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------
class QuerySet:
    """Contains object, and enables filtering or other operations. Is allowed to access object._."""
    def __init__(self, objects_l):
        self._objects_l = objects_l

    @property
    def objects_l(self):
        return self._objects_l

    def filter(self, field_index_or_name, field_value, condition="="):
        """
        Filter all objects who's field value matches field_value according to given condition.

        Arguments
        ---------
        field_index_or_name: field index or name. Can access children with tuple or list.
        field_value_or_values: value on which to be matched.
        condition: "=" (equality)
        condition: 'in' (include in string field)

        Returns
        -------
        QuerySet containing filtered objects.
        """
        if condition not in ("=", 'in'):
            raise IDFError("Unknown condition: '%s'." % condition)

        search_tuple = (field_index_or_name,) if isinstance(field_index_or_name, str) else field_index_or_name

        result_l = []
        for o in self._objects_l:
            current_value = o
            for level in search_tuple:
                current_value = current_value._.get_value(level)
            if condition == '=':
                if isinstance(current_value, str):
                    if current_value.lower() == field_value.lower():
                        result_l.append(o)
                else:
                    if current_value == field_value:
                        result_l.append(o)
            elif condition == 'in':
                if not isinstance(current_value, str):
                    raise IDFError("condition 'in' can not been performed on field_value  of type %s." % type(field_value))
                if field_value.lower() in current_value.lower():
                    result_l.append(o)
            else:
                raise IDFError("unknown condition : '%s'" % condition)

        return QuerySet(result_l)

    @property
    def one(self):
        """
        Checks that query set only contains one object and returns it.
        """
        if len(self._objects_l) == 0:
            raise ObjectDoesNotExist("Query set contains no value.")
        if len(self._objects_l) > 1:
            raise MultipleObjectsReturned("Query set contains more than one value.")
        return self[0]

    def __getitem__(self, item):
        return self._objects_l[item]

    def __iter__(self):
        return iter(self._objects_l)

    def __str__(self):
        return "<QuerySet: %s>" % str(self._objects_l)

    def __call__(self, object_descriptor_ref=None):
        """Returns all objects having given object descriptor ref (not case sensitive)."""
        if object_descriptor_ref is None:  # return a copy
            return QuerySet([o for o in self._objects_l])
        return QuerySet([o for o in self._objects_l if o._.ref.lower() == object_descriptor_ref.lower()])

    @check_cache_is_off
    def __add__(self, other):
        """
        Add new query set to query set (only new objects will be added).
        """
        self_set = set(self._objects_l)
        other_set = set(other.objects_l)
        intersect_set = self_set.intersection(other_set)
        new_objects_l = []
        new_objects_l.extend(self._objects_l)
        for idf_object in other.objects_l:
            if idf_object not in intersect_set:
                new_objects_l.append(idf_object)
        return QuerySet(new_objects_l)

    def __len__(self):
        return len(self._objects_l)


import os

import pandas as pd

from oplus.configuration import CONF


class EIOError(Exception):
    pass


class EndOfReportError(EIOError):
    pass


class EIO:
    def __init__(self, path, encoding=None):
        if not os.path.isfile(path):
            raise EIOError("No file at given path: '%s'." % path)
        self._path = path
        self._encoding = encoding

        self._tables_d = parse_eio(self._path, encoding=encoding)

    @property
    def table_refs(self):
        return self._tables_d.keys()

    def df(self, table_ref):
        return self._tables_d[table_ref].df

    def get_value(self, table_ref, column_name_or_i, filter_column_name_or_i, filter_criterion):
        if not table_ref in self._tables_d:
            raise EIOError("Unknown table_ref: '%s'." % table_ref)
        return self._tables_d[table_ref].get_value(column_name_or_i, filter_column_name_or_i, filter_criterion)


def parse_eio(path, encoding=None):
    headers_l2 = [["<Program Version>", "Program Version ID", "YMD"]]
    content_d = {}  # {ref: data_l2, ...}
    content_header_d = {}  # {name: header_row, ...}

    # _header_ref_pattern_ = "^([^<.]*)<([^>.]*)>(.*)$"

    with open(path, encoding=CONF.encoding if encoding is None else encoding) as f:
        for line_s in f:
            if line_s == "End of Data":
                break
            line_s = line_s.strip().strip(",")
            line_l = [c.strip() for c in line_s.split(",")]
            if line_s[0][0] == "!":  # header
                headers_l2.append([line_l[0][1:].strip()] + line_l[1:])
            else:  # content
                ref = line_l[0]
                if not ref in content_d:
                    content_d[ref] = []
                content_d[ref].append(line_l[1:])
                # find header if necessary
                if not ref in content_header_d:
                    for i in range(len(headers_l2)-1, -1, -1):
                        if ref.lower() in headers_l2[i][0].lower():
                            content_header_d[ref] = i
                            break

    # prepare data for dataframes
    tables_d = {}
    for ref, data in content_d.items():
        header_i = content_header_d.get(ref)
        if header_i is None:  # header not found
            header_columns = []
        else:
            header_columns = headers_l2[header_i][1:]
            if len(header_columns) == 1 and ("Months From Jan to Dec" in header_columns[0]):
                header_columns = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

        # prepare columns and data
        data_columns_nb = max([len(r) for r in data])
        columns_nb = max(data_columns_nb, len(header_columns))

        _columns = header_columns + ["c%i" % i for i in range(columns_nb-len(header_columns))]
        _data = []
        for r in data:
            _data.append(r + [None]*(columns_nb-len(r)))

        tables_d[ref] = EIOTable(ref, _columns, _data)

    return tables_d


class EIOTable:
    def __init__(self, ref, columns, data):
        """
        Stores dataframe info without parsing types (otherwise dtypes are changed even if dtyp='object' is asked...)
        """
        # check
        col_len = len(columns)
        for i, r in enumerate(data):
            if len(r) != col_len:
                raise EIOError("Wrong number of columns in row %i of table %s." % (i, ref))

        self._ref = ref
        self._columns = columns
        self._data = data

    @property
    def df(self):
        _df = pd.DataFrame(data=self._data, columns=self._columns, dtype="object")
        _df.name = self._ref
        return _df

    def get_value(self, column_name_or_i, filter_column_name_or_i, filter_criterion):
        """
        Returns first occurrence of value of filter column matching filter criterion.
        """
        # find column indexes
        column_i = self._get_column_index(column_name_or_i)
        filter_column_i = self._get_column_index(filter_column_name_or_i)

        filter_fct = {
            float: lambda x: float(x) == filter_criterion,
            int: lambda x: int(x) == filter_criterion,
            str: lambda x: str(x).lower() == filter_criterion.lower()
        }[type(filter_criterion)]

        for row_i, row in enumerate(self._data):
            if filter_fct(row[filter_column_i]):
                break
        else:
            raise EIOError("Filter did not return any values.")

        return self._data[row_i][column_i]

    def _get_column_index(self, column_name_or_i):
        if isinstance(column_name_or_i, int) or isinstance(column_name_or_i, float):
            return column_name_or_i
        if not column_name_or_i in self._columns:
            raise EIOError("Unknown column '%s' for table '%s'." % (column_name_or_i, df.index.name))
        return self._columns.index(column_name_or_i)


import os
import tempfile

from oplus.configuration import CONF
from oplus.idf import IDF


idf = IDF(os.path.join(CONF.eplus_base_dir_path, "ExampleFiles", "1ZoneEvapCooler.idf"))
zone = idf("Zone").filter("name", "Main Zone").one
building = idf("Building").filter("name", "Bldg").one

# check info
if True:
    print(idf.info(sort_by_group=False, detailed=True))

if True:
    print(zone.info(detailed=False))

# check to_str
if True:
    zone.head_comment = "Hello\n\n\nhello!!"
    zone.tail_comment = "Here is my tail comment\nwritten on several lines..."
    zone.field_comment(0, zone.field_comment(0) + " **modified with\nline\nbreaks**")
    print(zone.to_str(style="console"))  # idf, console

# check save_as
if True:
    # modify idf comment
    idf.comment = "I HAVE MODIFIED THE COMMENTS\n2 blank lines follow\n\n" + idf.comment

    # modify building
    building.head_comment = "MY BUILDING HEAD COMMENT"
    building.field_comment("terrain", building.field_comment("terrain") + " WITH MY NEW COMMENT")
    building.tail_comment = idf.comment

    f = tempfile.TemporaryFile("r+")
    idf.save_as(f)
    f.seek(0)
    print(f.read())



import unittest

from oplus.idd import *
from oplus.configuration import CONF


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.basicConfig(level=logging.INFO)


class IDDTest(unittest.TestCase):
    def test_current_idd(self):
        IDD(path=os.path.join(CONF.eplus_base_dir_path, "Energy+.idd"))

    def test_all_idds(self):
        for version, base_dir_path in CONF.eplus_available_versions.items():
            IDD(path=os.path.join(CONF.eplus_base_dir_path, "Energy+.idd"))
            logger.info("Idd version %s tested: ok." % str(version))



import unittest
import os
import tempfile
import io

from oplus.idf import IDF
from oplus.epw import EPW
from oplus.standard_output import StandardOutputFile
from oplus.simulation import simulate
from oplus.configuration import CONF


class OneZoneEvapCooler(unittest.TestCase):
    """
    Tested under EPlus 8.1.0 on Windows (Geoffroy).
    """
    idf = IDF(os.path.join(CONF.eplus_base_dir_path, "ExampleFiles", "1ZoneEvapCooler.idf"))
    epw = EPW(os.path.join(CONF.eplus_base_dir_path, "WeatherData", "USA_IL_Chicago-OHare.Intl.AP.725300_TMY3.epw"))

    def test_simulation_control(self):
        for (simulation_control, expected_environments) in (
                (None, {"RunPeriod", "SummerDesignDay", "WinterDesignDay"}),
                ("RunPeriods", {"RunPeriod"}),
                ("Sizing", {"SummerDesignDay", "WinterDesignDay"})
        ):
            with tempfile.TemporaryDirectory() as dir_path:
                s = simulate(self.idf, self.epw, dir_path, simulation_control=simulation_control)

                self.assertIsInstance(s.eso, StandardOutputFile)
                self.assertIsInstance(s.mtr, StandardOutputFile)
                self.assertEqual(set(s.eso.environments), expected_environments)
                self.assertEqual(set(s.mtr.environments), expected_environments)
                break

    def test_redirect_output(self):
        with tempfile.TemporaryDirectory() as dir_path:
            out_f, err_f = io.StringIO("hey"), io.StringIO("ho")

            s = simulate(
                self.idf,
                self.epw,
                dir_path,
                stdout=out_f,
                stderr=err_f
            )
            self.assertGreater(len(out_f.getvalue()), 0)
            self.assertGreater(len(err_f.getvalue()), 0)



from setuptools import setup
from oplus.version import version
import os

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()


setup(
    name='oplus',

    version=version,

    packages=['oplus'],

    author="Geoffroy d'Estaintot",

    author_email="geoffroy.destaintot@openergy.fr",

    description="A python package for working with Energy Plus",

    long_description=long_description,

    install_requires=[
        'pandas',
        'plotly'
        ],

    url='https://github.com/Openergy/oplus',

    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: French",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.4",
    ],

    keywords=['data', 'simulation'],

    package_data={'oplus': ['*.txt']},

    include_package_data=True
)



import tempfile
import os

from oplus.configuration import CONF
from oplus import simulate, EIO

MAX_TESTS_NB = 500
START_FILE_NUM = 0
DEBUG_SIMUL_DIR_PATH = None  # r"C:\Users\Geoffroy\Desktop\simul_dir"  # -> set to None to bypass


def check():
    """
    Tested under EPlus 8.1.0 on Windows (Geoffroy).
    """
    # !! CAN BE VERY LONG
    epw_path = os.path.join(CONF.eplus_base_dir_path, "WeatherData",
                            "USA_VA_Sterling-Washington.Dulles.Intl.AP.724030_TMY3.epw")
    idf_dir_path = os.path.join(CONF.eplus_base_dir_path, "ExampleFiles")
    test_num = 0
    for file_num, file_name in enumerate(os.listdir(idf_dir_path)):
        if file_num < START_FILE_NUM:
            continue
        base, ext = os.path.splitext(file_name)
        if ext == ".idf":
            with tempfile.TemporaryDirectory() as simulation_dir_path:
                s = simulate(os.path.join(idf_dir_path, file_name), epw_path,
                             simulation_dir_path if DEBUG_SIMUL_DIR_PATH is None else
                             DEBUG_SIMUL_DIR_PATH)
                if s.exists("eio"):
                    eio = EIO(s.path("eio"))  # raise error if problem
                    test_num += 1
        if test_num == MAX_TESTS_NB:
            break

if __name__ == "__main__":
    check()



"""
Standard Output File
"""
import datetime as dt
import logging
import os

import pandas as pd

from oplus.configuration import CONF
from oplus.util import EPlusDt, get_start_dt


logger = logging.getLogger(__name__)


class StandardOutputFileError(Exception):
    pass


class StandardOutputFile:
    DETAILED = "Detailed"
    TIME_STEP = "TimeStep"
    HOURLY = "Hourly"
    DAILY = "Daily"
    MONTHLY = "Monthly"
    RUN_PERIOD = "RunPeriod"  # !! is used as a time_step and a Environment

    SUMMER_DESIGN_DAY = "SummerDesignDay"
    WINTER_DESIGN_DAY = "WinterDesignDay"

    ENVIRONMENTS = (RUN_PERIOD, SUMMER_DESIGN_DAY, WINTER_DESIGN_DAY)
    TIME_STEPS = (DETAILED, TIME_STEP, HOURLY, DAILY, MONTHLY, RUN_PERIOD)

    def __init__(self, path, encoding=None, start=None):
        if not os.path.exists(path):
            raise StandardOutputFileError("No file at given path: '%s'." % path)
        self._path = path
        self._encoding = encoding

        self._start_dt = None if start is None else get_start_dt(start)

        # {run_period: {hourly: df, daily: df, monthly: df}, summer_design: df, winter_design: df}
        # datetime indexes are not managed here
        self._envs_d = self._parse()

    def set_start(self, start):
        self._start_dt = get_start_dt(start)

    def _parse(self):
        # todo: optimize (maybe readvarseso)
        with open(self._path, "r", encoding=CONF.encoding if self._encoding is None else self._encoding) as f:
            return parse_output(f)

    def df(self, environment=None, time_step=None, start=None, datetime_index=None):
        # todo: start does not always work
        """
        environment: 'RunPeriod', 'SummerDesignDay', 'WinterDesignDay' (default: first available)
        time_step: 'Detailed', 'TimeStep', 'Hourly', 'Daily', 'Monthly', 'RunPeriod' (default: first available)
        datetime_index: if None: True if possible, else False
        start: explain mechanism (can have been set with 'set_start' or while initialization)
        """
        # ------------------------------------ manage arguments --------------------------------------------------------
        # ENVIRONMENT
        # set environment if needed
        if environment is None:
            for environment in self.ENVIRONMENTS:
                if environment in self._envs_d:
                    break

        # check if environment is ok
        if not environment in self.ENVIRONMENTS:
            raise StandardOutputFileError("Unknown environment: '%s'." % environment)

        # check availability
        if not environment in self._envs_d:  # no available environment
            return None

        # set environment
        env_d = self._envs_d[environment]

        # TIME STEP
        # set time step if needed
        if time_step is None:
            for time_step in self.TIME_STEPS:
                if time_step in env_d:
                    break

        # check if time step is ok
        if not time_step in self.TIME_STEPS:
            raise StandardOutputFileError("Unknown time_step: '%s' (must be: %s)." %
                                          (time_step, ", ".join(self.TIME_STEPS)))

        # check availability
        if not time_step in env_d:  # no available time step
            return None

        # start_dt
        start_dt = self._start_dt if start is None else get_start_dt(start)

        # datetime_index
        if datetime_index is None:
            datetime_index = start_dt is not None
        if (datetime_index is True) and (start_dt is None):
            raise StandardOutputFileError("datetime_index mode can only be used if you indicated start. Use tuple "
                                          "index mode or registered start (set_start, on initialization of Output "
                                          "object or as df method argument.")

        # ------------------------------------ manage data frame -------------------------------------------------------
        # fetch dataframe
        if not time_step in env_d:
            return None
        df = env_d[time_step]

        # return if no conversion needed
        if not datetime_index or (time_step == self.RUN_PERIOD):
            return df

        # convert if needed
        start_esodt = EPlusDt.from_datetime(start_dt)

        if time_step in (self.DETAILED, self.TIME_STEP, self.HOURLY):
            row_to_esodt = lambda row: EPlusDt(*row[:4])
        elif time_step == self.DAILY:
            row_to_esodt = lambda row: EPlusDt(*(row[:2] + (1, 0)))
        else:  # monthly (RunPeriod has been returned)
            row_to_esodt = lambda row: EPlusDt(row, 1, 1, 0)

        start_standard_dt = start_esodt.standard_dt

        def row_to_dt(row):
            esodt = row_to_esodt(row)
            _year = start_dt.year + 1 if esodt.standard_dt <= start_standard_dt else start_dt.year
            return esodt.datetime(_year)

        df = df.copy()
        df.index = df.index.map(row_to_dt)
        df.sort(inplace=True)
        freq = None
        if time_step in (self.TIME_STEP, self.DETAILED):
            for year, year_df in df.groupby(lambda x: x.year):
                freq = year_df.index.inferred_freq
                if freq is not None:
                    break
            else:
                logger.warning("Could not find freq for sub-hourly data (not enough values). Did not reindex.")
                return df
        elif time_step == self.HOURLY:
            freq = "H"
        elif time_step == self.DAILY:
            freq = "D"
        elif time_step == self.MONTHLY:
            freq = "MS"

        # check that everything is ok while reindex
        before_nb = len(df)
        # todo: reindex will not work if a schedule is associated
        df = df.reindex(index=pd.date_range(df.index[0], df.index[-1], freq=freq))
        null_nb = (df.notnull().sum(axis=1) == 0).sum()
        if len(df) != before_nb + null_nb:
            logger.error(
                "BUG: Some values were lost during reindex (before reindex: %i, after: %i (%i full, %i empty)."
                % (before_nb, len(df), len(df)-null_nb, null_nb))

        return df

    def info(self):
        msg = ""
        for simulation_period in self._envs_d:
            msg += "\n\t%s: %s" % (simulation_period, ", ".join(self._envs_d[simulation_period].keys()))
        if msg == "":
            msg = "No data available."
        else:
            msg = "Available data:" + msg
        return msg

    @property
    def environments(self):
        return self._envs_d.keys()


def parse_output(file_like):
    """
    Only parses hourly (or infra-hourly) data, but does not raise Exception if there is daily or monthly data.
    Reporting frequencies 'Detailed' and 'RunPeriod' not implemented.
    """
    _detailed_ = "Detailed"
    _timestep_ = "TimeStep"
    _hourly_ = "Hourly"
    _daily_ = "Daily"
    _monthly_ = "Monthly"
    _run_period_ = "RunPeriod"

    # ----------------------- LOAD METERS
    # VERSION INFO
    next(file_like)

    # DATA DICTIONARY
    codes_d = {_detailed_: {}, _timestep_: {}, _hourly_: {}, _daily_: {}, _monthly_: {}, _run_period_: {}}
    # {interval:
    #       report_code: [(var_name, var_unit), ...], ...}
    while True:
        # separate content and comment
        line_s = next(file_like)
        line_l = line_s.split("!")
        if len(line_l) == 1:
            content_s, comment = line_l[0], ""
        else:
            content_s, comment = line_l
            comment = comment.strip()
        content_s = content_s.strip()

        if content_s == "End of Data Dictionary":
            break

        #report code
        content_l = content_s.split(",", 2)
        report_code, items_number, vars_l_s = int(content_l[0]), int(content_l[1]), content_l[2]

        # store information if interesting
        if report_code > 5:
            for (interval, pattern) in ((_detailed_, "Each Call"),
                                        (_timestep_, _timestep_),
                                        (_hourly_, _hourly_),
                                        (_daily_, _daily_),
                                        (_monthly_, _monthly_),
                                        (_run_period_, _run_period_)):
                if pattern in comment:
                    break
            else:
                raise StandardOutputFileError("Interval not found: '%s'" % line_s)
            codes_d[interval][report_code] = []
            for var_s in vars_l_s.split(",", items_number-1):
                try:
                    var_name, right_s = var_s.split("[")
                    var_unit = right_s[:-1].strip()
                except ValueError:
                    var_name, var_unit = var_s, None
                codes_d[interval][report_code] = (var_name.strip(), var_unit)

    # ------------------------ LOAD DATA
    # constants
    _begin_ = "begin"
    _data_d_ = "data_d"
    _index_l_ = "index_l"
    _dst_l_ = "dst_l"
    _day_types_l_ = "day_types_l"

    # shared local variables
    data_d, index_l, day_types_l, dst_l = None, None, None, None  # dataframe environments
    month_num, day_num, hour_num, end_minute_num, day_type, dst = None, None, None, None, None, None

    # global variables
    raw_envs_l = []

    # loop
    while True:
        line_s = next(file_like).strip()
        if line_s == "End of Data":
            break

        # find item num
        item_num_s, right_s = line_s.split(",", 1)
        item_num = int(item_num_s)

        if item_num > 5:  # value to store
            try:
                val = float(right_s)
            except ValueError:  # happens for 'RunPeriod' or 'Monthly' time step
                val = float(right_s.split(",")[0])  # we don't parse min and max

            # value has been parsed correctly
            try:
                data_d[item_num].append(val)  # data_d belongs to current env
            except (TypeError, KeyError):
                # data_d has not been initialized. Happens for first value of item_num 2, if interval is not known yet
                # find interval
                for interval in (_hourly_, _timestep_, _detailed_):
                    if item_num in codes_d[interval]:
                        break
                else:
                    raise StandardOutputFileError("Interval not found: '%s'." % line_s)

                # activate item
                data_d = raw_env_d[interval][_data_d_]
                index_l = raw_env_d[interval][_index_l_]
                day_types_l = raw_env_d[interval][_day_types_l_]
                dst_l = raw_env_d[interval][_dst_l_]

                # store
                index_l.append((month_num, day_num, hour_num, end_minute_num))
                day_types_l.append(day_type)
                dst_l.append(dst)

                # append as tried before
                data_d[item_num].append(val)

        elif item_num == 5:  # run period data
            # activate env
            data_d = raw_env_d[_run_period_][_data_d_]
            index_l, day_types_l, dst_l = None, None, None  # not used for run period env

        elif item_num == 2:  # hourly (or timestep) data
            # parse
            # 0-sim_day, 1-month_num, 2-day_num, 3-dst, 4-hour_num, 5-start_minute, 6-end_minute, 7-day_type
            # we don't use start_minute_num
            right_l = right_s.split(",")
            month_num = int(right_l[1])
            day_num = int(right_l[2])
            dst = int(right_l[3])
            hour_num = int(right_l[4])
            end_minute_num = int(float(right_l[6]))
            day_type = right_l[7]

            # deactivate everything (we don't know yet if detailed, hourly or timestep: must wait for first data)
            data_d, index_l, day_types_l, dst_l = None, None, None, None

        elif item_num == 3:  # daily
            # activate env
            data_d = raw_env_d[_daily_][_data_d_]
            index_l = raw_env_d[_daily_][_index_l_]
            dst_l = raw_env_d[_daily_][_dst_l_]
            day_types_l = raw_env_d[_daily_][_day_types_l_]

            # parse
            # 0-sim_day, 1-month_num, 2-day_num, 3-dst, 4-day_type
            right_l = right_s.split(",")
            month_num = int(right_l[1])
            day_num = int(right_l[2])
            dst = int(right_l[3])
            day_type = right_l[4]

            # store
            index_l.append((month_num, day_num))
            day_types_l.append(day_type)
            dst_l.append(dst)

        elif item_num == 4:  # monthly
            # activate env
            data_d = raw_env_d[_monthly_]["data_d"]
            index_l = raw_env_d[_monthly_]["index_l"]
            dst_l, day_types_l = None, None  # not used in monthly environment

            # parse
            # 0-sim_day, 1-month_num
            right_l = right_s.split(",")
            month_num = int(right_l[1])

            # store
            index_l.append(month_num)

        elif item_num == 1:  # new environment
            raw_env_d = {_begin_: line_s}
            for interval in (_detailed_, _timestep_, _hourly_, _daily_, _monthly_, _run_period_):  # meters and index
                if len(codes_d[interval]) == 0:  # no meters
                    continue
                raw_env_d[interval] = dict(data_d=dict([(k, []) for k in codes_d[interval]]), index_l=[])
                if interval == _run_period_:
                    raw_env_d[interval][_index_l_] = ["Total"]
                if interval in (_detailed_, _timestep_, _hourly_, _daily_):  # dst, day type
                    raw_env_d[interval][_dst_l_] = []
                    raw_env_d[interval][_day_types_l_] = []
            raw_envs_l.append(raw_env_d)

    # create environments and store
    envs_d = {}
    if len(raw_envs_l) == 1:
        env_names_l = ["RunPeriod"]
    elif len(raw_envs_l) == 2:
        env_names_l = ["SummerDesignDay", "WinterDesignDay"]
    elif len(raw_envs_l) == 3:
        env_names_l = ["SummerDesignDay", "WinterDesignDay", "RunPeriod"]
    else:
        logger.error("More than three environments were found, unhandled situation. Only first three will be used.")
        env_names_l = ["SummerDesignDay", "WinterDesignDay", "RunPeriod"]
    for env_name, raw_env_d in zip(env_names_l, raw_envs_l[:3]):
        # find env_name
        # for interval in (_detailed_, _timestep_, _hourly_, _daily_):
        #     if not interval in raw_env_d:
        #         continue
        #     first_day_type = raw_env_d[interval][_day_types_l_][0]
        #     env_name = first_day_type if first_day_type in ("SummerDesignDay", "WinterDesignDay") else "RunPeriod"
        #     break
        # else:  # monthly or runperiod
        #     env_name = "RunPeriod"
        #     logger = logging.getLogger(default_logger_name if logger_name is None else logger_name)
        #     logger.error("Did not find env_name for env: '%s'. Env has been skipped." % raw_env_d[_begin_])
        #     continue

        # create and store dataframes
        for interval in (_timestep_, _hourly_, _daily_, _monthly_, _run_period_):
            if interval not in raw_env_d:  # no data
                continue

            # CREATE
            # index
            index_l = raw_env_d[interval][_index_l_]
            if len(index_l) == 0:  # no data
                continue

            if interval in (_detailed_, _timestep_, _hourly_, _daily_):  # multi-index
                names = {
                    _detailed_: ["month", "day", "hour", "minute"],
                    _timestep_: ["month", "day", "hour", "minute"],
                    _hourly_: ["month", "day", "hour", "minute"],
                    _daily_: ["month", "day"]
                }[interval]
                index = pd.MultiIndex.from_tuples(index_l, names=names)
            else:
                index = pd.Index(index_l)

            # dataframe
            data_d = raw_env_d[interval][_data_d_]
            df = pd.DataFrame(data_d, index=index)
            df.rename(columns=dict([(k, codes_d[interval][k][0]) for k in codes_d[interval]]), inplace=True)

            # add dst and day_type if available
            if interval in (_timestep_, _hourly_, _daily_):
                df.insert(0, "dst", raw_env_d[interval]["dst_l"])
                df.insert(0, "day_type", raw_env_d[interval]["day_types_l"])

            # STORE
            if not env_name in envs_d:
                envs_d[env_name] = {}
            if interval in envs_d[env_name]:
                logger.error("Same environment has two identical time steps: '%s'." % raw_env_d[1])
            envs_d[env_name][interval] = df

    return envs_d


def __parse_out_optimized1(file_like):
    """
    Optimization failed...
    Only parses hourly (or infra-hourly) data, but does not raise Exception if there is daily or monthly data.
    Does not transform index into datetime
    """
    # ----------------------- LOAD METERS
    # VERSION INFO
    next(file_like)

    # DATA DICTIONARY
    report_codes_d = {}  # {report_code: [(var_name, var_unit), ...], ...}
    while True:
        line_s = next(file_like).split("!")[0].rstrip()  # we don't take comments and strip the newline character

        if line_s == "End of Data Dictionary":
            break
        line_l = line_s.split(",", 2)

        #report code
        report_code, items_number, vars_l_s = int(line_l[0]), int(line_l[1]), line_l[2]

        if report_code > 5:
            report_codes_d[report_code] = []
            for var_s in vars_l_s.split(",", items_number-1):
                try:
                    var_name, right_s = var_s.split("[")
                    var_unit = right_s[:-1].strip()
                except ValueError:
                    var_name, var_unit = var_s, None
                report_codes_d[report_code] = (var_name.strip(), var_unit)

    # ------------------------ LOAD DATA
    # we only use hourly (or sub-hourly data)
    data = []  # instant_id, timestep_id, meter_i, value
    instant = []  # instant_id, timestep_id, month_num, day_num, hour_num, end_minute_num
    # day_type_id (0 -> simulation, 1 -> SummerDesignDay, 2 -> WinterDesignDay)
    current_instant_id = -1

    day_types_d = {"SummerDesignDay": 1, "WinterDesignDay": 2}  # 0
    for line in file_like:
        line_l = line.split(",", 1)
        try:
            data.append((current_instant_id, int(line_l[0]), float(line_l[1])))
        except ValueError:
            try:
                # line_l: 0-sim_day, 1-month_num, 2-day_num, 3-dst, 4-hour_num, 5-start_minute,
                # 6-end_minute, 7-day_type
                # we don't use sim_day nor start_minute_num
                instant_l = line_l[1].split(",")
                instant.append(
                    (current_instant_id+1,  # instant_id
                     int(instant_l[1]),  # month_num
                     int(instant_l[2]),  # day_num
                     int(instant_l[4]),  # hour_num
                     int(float(instant_l[6])),  # end_minute_num
                     int(line_l[0]),  # timestep_id
                     day_types_d.get(instant_l[7].strip(), 0))  # day_type_id
                    # we didn't use instant_l[0] (sim_day), nor instant_l
                )
                current_instant_id += 1
            except ValueError:
                if int(line_l[0]) != 1:
                    raise Exception("Did not understand line: '%s'." % line)
            except IndexError:
                if line.strip() == 'End of Data':
                    break
    # manage index
    instant_df = pd.DataFrame(instant, columns=["instant_id", "month", "day", "hour", "minute", "timestep_id",
                                                "day_type_id"])
    instant_df["instant_id"] = instant_df.index
    instant_df = instant_df[instant_df["timestep_id"] == 2]
    instant_df = instant_df[instant_df["day_type_id"] == 0]

    instant_df = instant_df.set_index(["month", "day", "hour", "minute"])
    del instant_df["day_type_id"]
    del instant_df["timestep_id"]

    # reshape data and join
    data_df = pd.DataFrame(data, columns=["instant_id", "meter_id", "value"])
    data_df = data_df.pivot(index="instant_id", columns="meter_id", values="value")

    df = instant_df.join(data_df, on="instant_id", how="left")
    del df["instant_id"]
    df.rename(columns=dict([(k, report_codes_d[k][0]) for k in report_codes_d]), inplace=True)

    return df


if __name__ == "__main__":
    pass
    # import datetime as dt
    # _start_dt = dt.datetime.now()
    # out = StandardOutputFile(r"C:\Users\Geoffroy\Desktop\simul_dir\oplus.mtr")
    # print(out.info())
    # print(out.df(time_step="Monthly", datetime_index=True, start=1999))
    # print(dt.datetime.now()-_start_dt)